using Godot;

public class EnvironmentSystem
{
    private readonly TileGrid _tiles;
    private readonly float _diffusionCoefficient;

    private float[] _nextTemperatures = null;

    public EnvironmentSystem(TileGrid tiles, float diffusionCoefficient)
    {
        _tiles = tiles;
        _diffusionCoefficient = diffusionCoefficient;
        _nextTemperatures = new float[tiles.Count];
    }

    public void Process()
    {
        GenerateHeat();
        Diffuse();
    }

    private float TransferHeat(float t0, float alpha, float nx, float ny, float px, float py)
    {
        float d2tdx2 = nx - 2 * t0 + px;
        float d2tdy2 = ny - 2 * t0 + py;
        return t0 + alpha * (d2tdx2 + d2tdy2);
    }

    private void GetNeighbourTemperatures(int x, int y, out float nx, out float ny, out float px, out float py)
    {
        // default value
        var t = _tiles[x, y].temperature;

        nx = x > 0 ? _tiles[x - 1, y].temperature : t;
        px = x < _tiles.Size - 1 ? _tiles[x + 1, y].temperature : t;

        ny = y > 0 ? _tiles[x, y - 1].temperature : t;
        py = y < _tiles.Size - 1 ? _tiles[x, y + 1].temperature : t;
    }
    private void GenerateHeat()
    {
        for (int i = 0; i < _tiles.Count; i++)
        {
            var tile = _tiles[i];
            var type = tile.type;
            tile.temperature += type.HeatGeneration;
            ApplyHeatDamage(ref tile);
            _tiles[i] = tile;
        }
    }

    private void ApplyHeatDamage(ref Tile tile)
    {
        if (!(tile.type is IDamageable damageable))
            return;

        var delta = tile.temperature - damageable.Threshold;
        if (delta < 0)
        {
            // we want to heal based on how many wild tiles surround us
            float healRate = 0.01f;

            // count neighbours
        }
        else
        {
            // take damage
            // TODO: parameterised balancing
            tile.currentHealth -= delta;
            tile.currentHealth = Mathf.Clamp(tile.currentHealth, 0, 1);
        }

    }

    public int GetNeighbourCount<T>(int x, int y) where T : TileType
    {
        int count = 0;

        for (int i = -1; i < 1; i += 2)
        {
            for (int j = -1; j < 1; j += 2)
            {
                int xi = x + 1;
                int yj = y + 1;

                // assume tiles out side of bounds are the same type as this one
                if (!_tiles.IsInBounds(xi, yj))
                {
                    count += 1;
                    continue;
                }

                if (!(_tiles[xi, yj] is T))
                    continue;

                count++;
            }
        }

        return count;
    }

    private void Diffuse()
    {
        var D = _diffusionCoefficient;

        for (int i = 0; i < _tiles.Count; i++)
        {
            float t = _tiles[i].temperature;

            _tiles.MapIndex(i, out var x, out var y);

            GetNeighbourTemperatures(
                x, y,
                out var nx,
                out var ny,
                out var px,
                out var py);

            var temperature = TransferHeat(t, D, nx, ny, px, py);

            // TODO: what if it's really really cold out?
            temperature = Mathf.Max(0, temperature);
            _nextTemperatures[i] = temperature;
        }

        for (int i = 0; i < _tiles.Count; i++)
        {
            var tile = _tiles[i];
            tile.temperature = _nextTemperatures[i];
            _tiles[i] = tile;
        }
    }
}
